"""Classes and methods to collect information needed by Feed2Fedi to make posts on Fediverse instance."""

import os
import re
from pathlib import Path
from typing import Any
from urllib.parse import urlsplit

import feedparser
import httpx
from bs4 import BeautifulSoup
from httpx import AsyncClient
from httpx import HTTPError
from httpx import StreamError
from markdownify import markdownify


class FeedReader:
    """Instance hold feed items for RSS/Atom feed passed during instantiation."""

    def __init__(self, feed: str) -> None:
        with httpx.Client() as client:
            feed_text = client.get(url=feed).text
        self.items = feedparser.parse(feed_text).entries
        self.prepare_items_params()
        self.sort_entries()

    def sort_entries(self) -> None:
        """Sorts entries."""
        if all("published_parsed" in item for item in self.items) and all(
            item["published_parsed"] for item in self.items
        ):
            self.items.sort(key=lambda item: item["published_parsed"])

    def prepare_items_params(self) -> None:
        """Prepare items."""
        for item in self.items:
            item.params = {
                "title": item["title"] if "title" in item else None,
                "published": item["published"] if "published" in item else None,
                "updated": item["updated"] if "updated" in item else None,
                "link": item["link"] if "link" in item else None,
                "author": item["author"] if "author" in item else None,
            }

            if item.has_key("summary"):
                # Let's assume item content is HTML
                item.params["content_html"] = item["summary"]
                item.params["content_markdown"] = markdownify(item["summary"], strip=["img"]).strip()
                item.params["content_plaintext"] = markdownify(item["summary"], convert=["br", "p"]).strip()

    @staticmethod
    def determine_image_url(item: feedparser.FeedParserDict, image_selector: str) -> list[str]:
        """Determine URL for article image.

        :param item: Item to determine an image URL for
        :returns:
            List of strings with URL to article image
        """
        images: list[str] = []
        if item.has_key("summary"):
            parsed_content = BeautifulSoup(item.get("summary"), features="html.parser")
            images = [str(image.attrs["src"]) for image in parsed_content.select(image_selector)]

        if images:
            return images

        if item.has_key("description"):
            parsed_content = BeautifulSoup(item.get("description"), features="html.parser")
            images = [str(image.attrs["src"]) for image in parsed_content.select(image_selector)]

        if item.has_key("link"):
            images = [link.get("href") for link in item.links if "image" in link.get("type")]

        if images:
            return images

        if image_url := item.get("media_thumbnail", [{}])[0].get("url"):
            images = [image_url]

        if images:
            return images

        if image_url := item.get("media_content", [{}])[0].get("url"):
            images = [image_url]

        return images


async def get_file(
    img_url: str,
    file: Any,
    supported_mime_types: list[str],
) -> str | None:
    """Save a file located at img_url to a file located at filepath.

    :param img_url: url of imgur image to download
    :param file: File to write image to
    :param supported_mime_types: List of strings representing mime types supported by the instance server

    :returns:
        mime_type (string): mimetype as returned from URL
    """
    mime_type = await determine_mime_type(img_url=img_url)

    try:
        if not mime_type or (mime_type not in supported_mime_types):
            return None

        async with AsyncClient(http2=True) as client:
            async with client.stream(method="GET", url=img_url) as response:
                response.raise_for_status()
                async for data_chunk in response.aiter_bytes():
                    file.write(data_chunk)

        return mime_type

    except (HTTPError, StreamError) as save_image_error:
        print(f"collect.py - get_file(...) -> None - download failed with: {save_image_error}")

    return None


async def determine_mime_type(img_url: str) -> str | None:
    """Determine suitable filename for an image based on URL.

    :param img_url: URL to image to determine a file name for.
    :returns:
        mime-type in a String or None
    """
    # First check if URL starts with http:// or https://
    regex = r"^https?://"
    match = re.search(regex, img_url, flags=0)
    if not match:
        print(f"Post link is not a full link: {img_url}")
        return None

    # Acceptable image formats
    image_formats = (
        "image/png",
        "image/jpeg",
        "image/gif",
        "image/webp",
        "video/mp4",
    )

    file_name = Path(os.path.basename(urlsplit(img_url).path))

    # Determine mime type of linked media
    try:
        async with AsyncClient(http2=True) as client:
            response = await client.head(url=img_url)
            response.raise_for_status()
            headers = response.headers
            content_type = headers.get("content-type", None)

    except HTTPError as error:
        print(f"Error while opening URL: {error}")
        return None

    if content_type in image_formats:
        return str(content_type)

    if content_type == "application/octet-stream" and file_name.suffix == ".webp":
        return "image/webp"

    print(f"URL does not point to a valid image file: {img_url}")
    return None



"""Classes and methods needed to publish posts on a Fediverse instance."""

import asyncio
import re
import tempfile
import traceback
from pathlib import Path

import puremagic
import yt_dlp
import yt_dlp.utils
from feedparser import FeedParserDict
from httpx import AsyncClient
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ClientError
from minimal_activitypub.client_2_server import NetworkError
from minimal_activitypub.client_2_server import RatelimitError
from stamina import retry
from stamina import retry_context
from whenever import ZonedDateTime

from feed2fedi.collect import FeedReader
from feed2fedi.collect import get_file
from feed2fedi.control import Actions
from feed2fedi.control import Configuration
from feed2fedi.control import FeedInfo
from feed2fedi.control import IgnoringLogger
from feed2fedi.control import PostRecorder


class Fediverse:
    """Helper class to publish posts on a fediverse instance from rss feed items."""

    def __init__(self, config: Configuration, post_recorder: PostRecorder) -> None:
        self.config = config
        self.post_recorder = post_recorder

    async def publish(
        self,
        items: list[FeedParserDict],
        feed: FeedInfo,
        post_limit: int | None,
        client: AsyncClient,
    ) -> int:
        """Publish posts to fediverse instance from content in the items list.

        :param post_limit: Optional; Number of statuses to post before returning
        :param items: Rss feed items to post
        :param feed: Section of config for current feed
        :param client: httpx AsyncClient to use for posting to fedi instance.

        :returns int: Number of new statuses posted.
        """
        statuses_posted = 0
        fediverse = ActivityPub(
            instance=self.config.fedi_instance,
            client=client,
            access_token=self.config.fedi_access_token,
        )
        for attempt in retry_context(on=NetworkError, attempts=3):
            with attempt:
                await fediverse.determine_instance_type()
                await fediverse.verify_credentials()

        if isinstance(feed.max_attachments, int):
            max_media = min(fediverse.max_attachments, feed.max_attachments)
        else:
            max_media = fediverse.max_attachments

        for item in items:
            if await self.post_recorder.duplicate_check(identifier=item.link):
                continue

            filter_action_drop, sensitive, spoiler_text = await self._apply_filters(feed, item)

            if filter_action_drop:
                continue

            status = ""
            if feed.prefix:
                status += f"{feed.prefix} - "
            status += self.config.bot_post_template.format(**item.params).replace("\\n", "\n")

            try:
                media_ids = await self._post_media(fediverse, feed, item, max_media)

                posted_status = await self._post_actual_status(fediverse, media_ids, sensitive, spoiler_text, status)

                print(f"Posted {posted_status['content']} to Fediverse at\n{posted_status['url']}")

                await self.post_recorder.log_post(shared_url=item.link)

                statuses_posted += 1
                if post_limit and statuses_posted >= post_limit:
                    break

            except RatelimitError:
                reset = fediverse.ratelimit_reset
                seconds = reset.timestamp() - ZonedDateTime.now("UTC").timestamp()
                print(
                    f'!!! Server "cool down" - waiting until {reset:%Y-%m-%d %H:%M:%S %z} '
                    f"({round(number=seconds)} seconds)"
                )
                await asyncio.sleep(delay=seconds)

            except ClientError as error:
                print(f"!!! Encountered error: {error}")
                traceback.print_tb(error.__traceback__)
                print("\nLog article to avoid repeat of error")
                await self.post_recorder.log_post(shared_url=item.link)

        return statuses_posted

    @retry(on=NetworkError, attempts=3)
    async def _post_actual_status(self, fediverse, media_ids, sensitive, spoiler_text, status):
        """Post actual status. This has been refactored into its own method to be able to apply retry logic."""
        posted_status = await fediverse.post_status(
            status=status,
            visibility=self.config.bot_post_visibility,
            media_ids=media_ids,
            sensitive=sensitive,
            spoiler_text=spoiler_text,
        )
        return posted_status

    async def _post_media(self, fediverse, feed, item, max_media) -> list[str] | None:
        """Post media if configured to do so."""
        media_ids: list[str] | None = None
        if self.config.bot_post_media and max_media:
            media_ids = await Fediverse._post_video(
                fediverse=fediverse,
                item=item,
                post_videos=feed.post_videos,
            )

            if not media_ids:
                media_ids = await Fediverse._post_images(
                    fediverse=fediverse,
                    item=item,
                    max_images=max_media,
                    image_selector=self.config.bot_post_image_selector,
                    supported_mime_types=fediverse.supported_mime_types,
                )
        return media_ids

    @staticmethod
    async def _apply_filters(feed, item) -> tuple[bool, bool, str | None]:
        """Apply filters to item."""
        filter_action_drop = False
        sensitive = False
        spoiler_text = None
        for item_filter in feed.filters:
            if item_filter.do_check(item):
                if item_filter.action == Actions.DROP:
                    filter_action_drop = True
                    continue
                elif item_filter.action == Actions.SEARCH_REPLACE:
                    search = item_filter.action_params["search"]
                    replace = item_filter.action_params["replace"]
                    item.params["content_html"] = re.sub(search, replace, item.params["content_html"])
                    item.params["content_markdown"] = re.sub(search, replace, item.params["content_markdown"])
                    item.params["content_plaintext"] = re.sub(search, replace, item.params["content_plaintext"])
                elif item_filter.action == Actions.MARK_CW:
                    sensitive = True
                    spoiler_text = item_filter.action_params
        return filter_action_drop, sensitive, spoiler_text

    @staticmethod
    @retry(on=NetworkError, attempts=3)
    async def _post_video(
        fediverse: ActivityPub,
        item: FeedParserDict,
        post_videos: bool,
    ) -> list[str]:
        """Post media to fediverse instance and return media ID.

        :param fediverse: ActivityPub api instance
        :param item: Feed item to load media from
        :param post_videos: Boolean indicating whether videos should be posted.
        :returns:
            List containing no, one  or multiple strings of the media id after upload
        """
        media_ids: list[str] = []
        filenames: list[str] = []

        if not post_videos:
            return media_ids

        try:
            filenames = await Fediverse._get_video(item)
        except yt_dlp.utils.DownloadError:
            # Skip and go to next processing type
            pass

        if len(filenames) > 0:
            for filename in filenames:
                magic_info = puremagic.magic_file(filename=filename)
                mime_type = magic_info[0].mime_type
                try:
                    if mime_type:
                        with Path(filename).open(mode="rb") as file:
                            media = await fediverse.post_media(file=file, mime_type=mime_type)
                        media_ids.append(media.get("id"))
                except ClientError as error:
                    raise ClientError from error
                finally:
                    Path(filename).unlink()

        return media_ids

    @staticmethod
    async def _get_video(item) -> list[str]:
        """Download videos."""
        filenames: list[str] = []
        ydl_opts = {
            "quiet": "true",
            "logger": IgnoringLogger(),
            "no_warnings": "true",
            "ignoreerrors": "true",
            "outtmpl": "%(id)s.%(ext)s",
        }
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            info = ydl.extract_info(item.link, download=True)
            ydl_info = ydl.sanitize_info(info)
            if ydl_info and ydl_info.get("requested_downloads"):
                for dl in ydl_info.get("requested_downloads"):
                    if filename := dl.get("filepath"):
                        filenames.append(filename)
        return filenames

    @staticmethod
    @retry(on=NetworkError, attempts=3)
    async def _post_images(
        fediverse: ActivityPub,
        item: FeedParserDict,
        image_selector: str,
        supported_mime_types: list[str],
        max_images: int = 1,
    ) -> list[str]:
        """Post media to fediverse instance and return media ID.

        :param fediverse: ActivityPub api instance
        :param item: Feed item to load media from
        :param max_images: number of images to post. Defaults to 1
        :param supported_mime_types:  List of strings representing mime types supported by the instance server

        :returns:
            List containing no, one  or multiple strings of the media id after upload
        """
        media_ids: list[str] = []
        media_urls = FeedReader.determine_image_url(item, image_selector)

        for url in media_urls:
            if len(media_ids) == max_images:
                break

            with tempfile.TemporaryFile() as temp_image_file:
                mime_type = await get_file(img_url=url, file=temp_image_file, supported_mime_types=supported_mime_types)
                if mime_type:
                    temp_image_file.seek(0)
                    media = await fediverse.post_media(
                        file=temp_image_file,
                        mime_type=mime_type,
                    )
                    media_ids.append(media["id"])

        return media_ids



"""Main processing and main entry point methods for Feed2Fedi."""

import asyncio
import sys
import traceback
from pathlib import Path
from typing import Annotated

import typer
from httpx import AsyncClient
from minimal_activitypub.client_2_server import ActivityPubError

from feed2fedi import DISPLAY_NAME
from feed2fedi import __version__
from feed2fedi.collect import FeedReader
from feed2fedi.control import PostRecorder
from feed2fedi.control import load_config
from feed2fedi.publish import Fediverse


async def main(config_file: Path, post_limit: int | None = None):
    """Read configuration and feeds, then make posts while avoiding duplicates.

    :param post_limit: Optional; number of new statuses to post before exiting. Default is "None" and if so we process
        all entries in all configured RSS/Atom feeds.
    :param config_file: Path and file name of file to use for reading and storing configuration from
    """
    error_encountered = False

    print(f"Welcome to {DISPLAY_NAME} {__version__}")

    config = await load_config(config_file_path=Path(config_file))

    async with (
        PostRecorder(history_db_path=config.cache_db_path) as post_recorder,
        AsyncClient(http2=True, timeout=30) as client,
    ):
        await post_recorder.prune(max_age_in_days=config.cache_max_age)

        fediverse = Fediverse(config=config, post_recorder=post_recorder)

        statuses_posted = 0
        for feed in config.feeds:
            items = FeedReader(feed=feed.url).items

            try:
                statuses_posted += await fediverse.publish(items=items, feed=feed, post_limit=post_limit, client=client)
            except ActivityPubError as publishing_error:
                error_encountered = True
                print(f"Encountered the following error during publishing feed items:\n{publishing_error}")
                traceback.print_tb(publishing_error.__traceback__)
                break

            if post_limit and statuses_posted >= post_limit:
                break

    if error_encountered:
        sys.exit(1)


async def import_urls(config_file: Path, url_file: Path) -> None:
    """Start import of URLS into cache db.

    :param config_file: Path and file name of file to use for reading and storing configuration from
    :param url_file: Path and file name to file to be imported
    """
    print(f"Welcome to {DISPLAY_NAME} {__version__}")
    print("\nImporting URLS into cache db from ...")

    config = await load_config(config_file_path=Path(config_file))

    post_recorder = PostRecorder(history_db_path=config.cache_db_path)
    await post_recorder.db_init()

    await post_recorder.import_urls(url_file=Path(url_file))

    await post_recorder.close_db()


def start_main() -> None:
    """Start processing, i.e. main entry point."""
    try:
        typer.run(start_main_shim)
    except asyncio.CancelledError:
        pass


def start_main_shim(
    config_file: Annotated[
        Path,
        typer.Option(
            "--config-file",
            "-c",
            file_okay=True,
            dir_okay=False,
            writable=True,
            resolve_path=True,
            help="filename and optional path to config file",
        ),
    ] = Path("./config.json"),
    post_limit: Annotated[
        int | None,
        typer.Option(
            "--limit",
            "-l",
            help="Limit how many new statuses should be posted",
        ),
    ] = None,
) -> None:
    """Start Feed2Fedi."""
    asyncio.run(main(config_file=config_file, post_limit=post_limit))


def start_import_shim(
    url_file: Annotated[
        Path,
        typer.Option(
            "--url-file",
            help="Path and file name to file to be imported. It should contain one URL per line",
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    config_file: Annotated[
        Path,
        typer.Option(
            "--config-file",
            "-c",
            file_okay=True,
            dir_okay=False,
            writable=True,
            resolve_path=True,
            help="filename and optional path to config file",
        ),
    ] = Path("./config.json"),
) -> None:
    """Import of URLS into cache db."""
    asyncio.run(import_urls(config_file=config_file, url_file=url_file))


def start_import() -> None:
    """Call start import shim for typer."""
    typer.run(start_import_shim)



"""Classes and methods to control how Feed2Fedi works."""

import re
import sys
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Final
from typing import TypeVar

import aiosqlite
import msgspec
import msgspec.json
from httpx import AsyncClient
from minimal_activitypub import Visibility
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError
from whenever import Instant
from whenever import ZonedDateTime
from whenever import days

from feed2fedi import DISPLAY_NAME
from feed2fedi import FILE_ENCODING
from feed2fedi import POST_RECORDER_SQLITE_DB
from feed2fedi import WEBSITE

CACHE_DB_PATH_DEFAULT: Final[str] = "./cache.sqlite"
CACHE_MAX_AGE_DEFAULT_30_DAYS: Final[int] = 30
BOT_POST_TEMPLATE_DEFAULT: Final[str] = r"{title}\n\n{link}"
BOT_POST_IMAGE_SELECTOR_DEFAULT: Final[str] = "img[src]"

ConfigClass = TypeVar("ConfigClass", bound="Configuration")
PR = TypeVar("PR", bound="PostRecorder")


class Checks(str, Enum):
    """Supported filters checks."""

    REGEX = "regex"
    ANY = "any"
    NONE = "none"


class Actions(str, Enum):
    """Supported filters actions."""

    DROP = "drop"
    SEARCH_REPLACE = "search_replace"
    MARK_CW = "mark_cw"
    NONE = "none"


class ItemsFilter(msgspec.Struct):
    """Defines ItemsFilters and actions."""

    check: str = Checks.NONE
    check_params: Any = None
    action: str = Actions.NONE
    action_params: Any = None

    def do_check(self, item) -> bool:
        """Check if ItemFilter `check` is valid."""
        if self.check == Checks.REGEX:
            if not isinstance(self.check_params, str):
                raise Exception("Regex filter should have string as check_params")
            return bool(re.match(self.check_params, item["summary"], re.DOTALL))
        elif self.check == Checks.ANY:
            return True
        elif self.check == Checks.NONE:
            return False
        else:
            raise Exception(f"Unknown check {self.check}")


class FeedInfo(msgspec.Struct):
    """Dataclass to hold info needed for each feed."""

    url: str
    prefix: str | None = None
    max_attachments: int | str = "max"
    filters: list[ItemsFilter] = []
    post_videos: bool = True


class Configuration(msgspec.Struct):
    """Dataclass to hold configuration settings for Feed2Fedi."""

    feeds: list[FeedInfo]
    fedi_instance: str
    fedi_access_token: str
    cache_max_age: int
    cache_db_path: str
    bot_post_media: bool
    bot_post_visibility: Visibility
    bot_post_template: str
    bot_post_image_selector: str = BOT_POST_IMAGE_SELECTOR_DEFAULT

    @classmethod
    async def create_default_config_file(cls, config_file_path: Path) -> None:
        """Create new configuration and save as new config file."""
        instance = Configuration._get_instance()
        access_token = await Configuration._get_access_token(instance=instance)
        new_feed = FeedInfo(url="http://feedparser.org/docs/examples/rss20.xml", prefix=None, max_attachments="max")
        new_config = Configuration(
            feeds=[new_feed],
            fedi_instance=instance,
            fedi_access_token=access_token,
            cache_max_age=CACHE_MAX_AGE_DEFAULT_30_DAYS,
            cache_db_path=CACHE_DB_PATH_DEFAULT,
            bot_post_media=True,
            bot_post_image_selector=BOT_POST_IMAGE_SELECTOR_DEFAULT,
            bot_post_visibility=Visibility.PUBLIC,
            bot_post_template=BOT_POST_TEMPLATE_DEFAULT,
        )
        json_config = msgspec.json.encode(new_config)
        with config_file_path.open(mode="wb") as config_file:
            config_file.write(msgspec.json.format(json_config, indent=4))

    @staticmethod
    def _get_instance() -> str:
        """Get instance URL from user.

        :returns:
            URL of fediverse instance.
        """
        instance = input("[...] Please enter the URL for the instance to connect to: ")
        return instance

    @staticmethod
    async def _get_access_token(instance: str) -> str:
        """Get access token from fediverse instance.

        :param instance: URL to fediverse instance
        :returns:
            Access token
        """
        try:
            async with AsyncClient(http2=True) as client:
                client_id, client_secret = await ActivityPub.create_app(
                    instance_url=instance,
                    client=client,
                    user_agent=DISPLAY_NAME,
                    client_website=WEBSITE,
                )

                # Get Authorization Code / URL
                authorization_request_url = await ActivityPub.generate_authorization_url(
                    instance_url=instance,
                    client_id=client_id,
                    user_agent=DISPLAY_NAME,
                )
                print(f"Please go to the following URL and follow the instructions:\n{authorization_request_url}")
                authorization_code = input("[...] Please enter the authorization code: ")

                # Validate authorization code and get access token
                access_token = await ActivityPub.validate_authorization_code(
                    client=client,
                    instance_url=instance,
                    authorization_code=authorization_code,
                    client_id=client_id,
                    client_secret=client_secret,
                )

        except ActivityPubError as error:
            print(f"! Error when setting up Fediverse connection: {error}")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

        return str(access_token)


class PostRecorder:
    """Record posts, check for duplicates, and deletes old records of posts."""

    LAST_POST_TS: Final[str] = "last-post-timestamp"

    def __init__(self: PR, history_db_path: str = f"./{POST_RECORDER_SQLITE_DB}") -> None:
        """Initialise PostRecord instance.

        :param history_db_path: Location where history db should be stored. Default to current directory (.)
        """
        self.history_db_file = history_db_path
        self.history_db: aiosqlite.Connection | None = None

    async def db_init(self: PR) -> None:
        """Initialise DB connection and tables if necessary."""
        self.history_db = await aiosqlite.connect(database=self.history_db_file)
        # Make sure DB tables exist
        await self.history_db.execute(
            "CREATE TABLE IF NOT EXISTS share (url TEXT PRIMARY KEY, shared_ts INT) WITHOUT ROWID"
        )
        await self.history_db.execute("CREATE INDEX IF NOT EXISTS index_ts ON share(shared_ts)")
        await self.history_db.execute("CREATE TABLE IF NOT EXISTS settings (key TEXT PRIMARY KEY, value) WITHOUT ROWID")
        await self.history_db.execute(
            "INSERT OR IGNORE INTO settings (key, value) VALUES (:key, :value)",
            {"key": PostRecorder.LAST_POST_TS, "value": 0},
        )
        await self.history_db.commit()

    async def duplicate_check(self: PR, identifier: str) -> bool:
        """Check identifier can be found in log file of content posted to
        Mastodon.

        :param identifier:
                Any identifier we want to make sure has not already been posted.
                This can be id of reddit post, url of media attachment file to be
                posted, or checksum of media attachment file.

        :returns:
                False if "identifier" is not in log of content already posted to
                Mastodon
                True if "identifier" has been found in log of content.
        """
        if self.history_db is None:
            raise AssertionError("Have you called db_init() first?")

        # check for Shared URL
        cursor = await self.history_db.execute("SELECT * FROM share where url=:url", {"url": identifier})
        if await cursor.fetchone():
            return True

        return False

    async def log_post(
        self: PR,
        shared_url: str,
    ) -> None:
        """Log details about posts that have been published.

        :param shared_url:
                URL of feed item
        """
        if self.history_db is None:
            raise AssertionError("Have you called db_init() first?")

        timestamp = Instant.now().timestamp()
        await self.history_db.execute(
            "INSERT INTO share (url, shared_ts) VALUES (?, ?)",
            (
                shared_url,
                timestamp,
            ),
        )
        await self.history_db.commit()

    async def get_setting(
        self: PR,
        key: str,
    ) -> Any:
        """Retrieve a setting from database.

        :param key: Key to setting stored in DB

        :return: Value of setting. This could be an int, str, or float
        """
        if self.history_db is None:
            raise AssertionError("Have you called db_init() first?")

        cursor = await self.history_db.execute(
            "SELECT value FROM settings WHERE key=:key",
            {"key": key},
        )
        row = await cursor.fetchone()
        if row is None:
            return None

        return row[0]

    async def save_setting(
        self: PR,
        key: str,
        value: int | str | float,
    ) -> None:
        """Save a setting to database.

        :param key: Key to setting stored in DB
        :param value: Value to store as a setting

        :return: None
        """
        if self.history_db is None:
            raise AssertionError("Have you called db_init() first?")

        await self.history_db.execute(
            "INSERT OR REPLACE INTO settings (key, value) VALUES (:key, :value)",
            {"key": key, "value": value},
        )

        await self.history_db.commit()

    async def prune(self, max_age_in_days: int) -> None:
        """Prune entries from db that are older than max_age_in_days.

        :param max_age_in_days: Maximum age of records to keep in DB
        """
        if self.history_db is None:
            raise AssertionError("Have you called db_init() first?")

        max_age_ts = (ZonedDateTime.now("UTC") - days(max_age_in_days)).timestamp()
        await self.history_db.execute(
            "DELETE FROM share WHERE shared_ts<:max_age_ts",
            {"max_age_ts": max_age_ts},
        )
        await self.history_db.commit()

    async def close_db(self: PR) -> None:
        """Close db connection."""
        if self.history_db:
            await self.history_db.close()

    async def import_urls(self, url_file: Path) -> None:
        """Import URLS from for example the cache.db of feed2toot.

        :param url_file: File path containing one URL per line
        """
        if self.history_db is None:
            raise AssertionError("Have you called db_init() first?")

        line_count = 0
        with url_file.open(mode="r", encoding=FILE_ENCODING) as import_file:
            while entry := import_file.readline():
                await self.log_post(shared_url=entry)
                line_count += 1

        print(f"Imported {line_count} urls")

    async def __aenter__(self):
        """Magic method to enable the use of an 'async with PostRecoder(...) as ...' block
        Ready the cache db for.
        """
        await self.db_init()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Magic method defining what happens when 'async with ...' block finishes.
        Close cache db.
        """
        await self.close_db()


class IgnoringLogger:
    """Logger class that ignores all logging silently."""

    def debug(self, msg):
        """Process debug log messages."""
        # For compatibility with youtube-dl, both debug and info are passed into debug
        # You can distinguish them by the prefix '[debug] '
        if msg.startswith("[debug] "):
            pass
        else:
            self.info(msg)

    def info(self, msg):
        """Process info log messages."""
        pass

    def warning(self, msg):
        """Process warning log messages."""
        pass

    def error(self, msg):
        """Process error log messages."""
        pass


async def load_config(config_file_path: Path) -> Configuration:
    """Load configuration values from file and create Configuration instance.

    :param config_file_path: File name to load configuration values from
    :returns:
        Configuration instance with values loaded from file_name
    """
    if not config_file_path.exists():
        await Configuration.create_default_config_file(config_file_path)

    with config_file_path.open(mode="rb") as config_file:
        new_config: Configuration = msgspec.json.decode(config_file.read(), type=Configuration)

    if isinstance(new_config.bot_post_visibility, str):
        new_config.bot_post_visibility = Visibility(new_config.bot_post_visibility)

    return new_config



"""Module specific values and logic."""

from importlib.metadata import version
from typing import Final

__version__: Final[str] = version(str(__package__))

DISPLAY_NAME: Final[str] = "Feed2Fedi"
WEBSITE: Final[str] = "https://codeberg.org/MarvinsMastodonTools/feed2fedi"

POST_RECORDER_SQLITE_DB: Final[str] = "cache.sqlite"
FILE_ENCODING: Final[str] = "UTF-8"



"""Module containing utilities and helper methods."""

import configparser
from itertools import zip_longest
from pathlib import Path
from typing import Annotated

import msgspec
import msgspec.json
import typer
from minimal_activitypub import Visibility

from feed2fedi.control import BOT_POST_IMAGE_SELECTOR_DEFAULT
from feed2fedi.control import Configuration
from feed2fedi.control import FeedInfo


def convert_config_json(  # nocl
    config_file: Annotated[
        Path,
        typer.Option(
            "--config-file",
            "-c",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
            help="filename and optional path to old config INI file",
        ),
    ],
    config_json: Annotated[
        Path,
        typer.Option(
            "--config-json",
            "-j",
            file_okay=True,
            dir_okay=False,
            writable=True,
            resolve_path=True,
            help="filename and optional path to new config JSON file",
        ),
    ],
) -> None:
    """Convert a .conf file to a .json config file."""
    config_ini = configparser.ConfigParser()
    with config_file.open(mode="r", encoding="UTF-8") as old_config_file:
        config_ini.read_file(f=old_config_file)

    old_feeds = [feed for _key, feed in config_ini.items(section="Feeds")]
    old_prefixes = [prefix for _key, prefix in config_ini.items(section="Prefixes")]

    new_feeds: list[FeedInfo] = []
    for feed, prefix in zip_longest(old_feeds, old_prefixes, fillvalue=""):
        new_feeds.append(FeedInfo(url=feed, prefix=prefix, max_attachments=1))

    new_config = Configuration(
        bot_post_visibility=Visibility(config_ini.get(section="Bot", option="post_visibility")),
        bot_post_media=bool(config_ini.get(section="Bot", option="post_media")),
        bot_post_image_selector=BOT_POST_IMAGE_SELECTOR_DEFAULT,
        bot_post_template="{title}\n\n{link}",
        cache_max_age=int(config_ini.get(section="Cache", option="max-age")),
        cache_db_path=config_ini.get(section="Cache", option="db-path"),
        fedi_instance=config_ini.get(section="Fediverse", option="instance"),
        fedi_access_token=config_ini.get(section="Fediverse", option="access-token"),
        feeds=new_feeds,
    )

    with config_json.open(mode="wb") as new_config_file:
        json_config = msgspec.json.encode(new_config)
        new_config_file.write(msgspec.json.format(json_config, indent=4))


def start_conversion() -> None:
    """Start the conversion process."""
    typer.run(convert_config_json)
